@@ -16,9 +16,9 @@
 
 import base64
 import copy
+import json
 import os
 import sys
-import uuid
 
 import bson
 from bson.raw_bson import RawBSONDocument
@@ -30,9 +30,11 @@ from bson.son import SON
 
 sys.path[0:0] = [""]
 
+import pymongocrypt.mongocrypt
+from pymongo_auth_aws.auth import AwsCredential
 from pymongocrypt.auto_encrypter import AutoEncrypter
 from pymongocrypt.binding import lib
-from pymongocrypt.compat import unicode_type, safe_bytearray_or_base64, PY3
+from pymongocrypt.compat import unicode_type, PY3
 from pymongocrypt.errors import MongoCryptError
 from pymongocrypt.explicit_encrypter import ExplicitEncrypter
 from pymongocrypt.mongocrypt import (MongoCrypt,
@@ -41,7 +43,11 @@ from pymongocrypt.mongocrypt import (MongoCrypt,
                                      MongoCryptOptions)
 from pymongocrypt.state_machine import MongoCryptCallback
 
-from test import unittest
+from test import unittest, mock
+
+import requests.exceptions
+import requests_mock
+
 
 # Data for testing libbmongocrypt binding.
 DATA_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), 'data'))
@@ -86,6 +92,7 @@ class TestMongoCryptOptions(unittest.TestCase):
         schema_map = bson_data('schema-map.json')
         valid = [
             ({'local': {'key': b'1' * 96}}, None),
+            ({ 'aws' : {} }, schema_map),
             ({'aws': {'accessKeyId': '', 'secretAccessKey': ''}}, schema_map),
             ({'aws': {'accessKeyId': 'foo', 'secretAccessKey': 'foo'}}, None),
             ({'aws': {'accessKeyId': 'foo', 'secretAccessKey': 'foo',
@@ -94,6 +101,9 @@ class TestMongoCryptOptions(unittest.TestCase):
               'local': {'key': b'1' * 96}}, None),
             ({'local': {'key': to_base64(b'1' * 96)}}, None),
             ({'local': {'key': Binary(b'1' * 96)}}, None),
+            ({'azure': {}}, None),
+            ({'azure': {'clientId': 'foo', 'clientSecret': 'bar'}}, None),
+            ({'gcp': {}}, None),
             ({'gcp': {'email': 'foo@bar.baz',
                       'privateKey': b'1'}}, None),
             ({'gcp': {'email': 'foo@bar.baz',
@@ -122,16 +132,16 @@ class TestMongoCryptOptions(unittest.TestCase):
                 {'aws': {'accessKeyId': 'foo'}},
                 {'aws': {'secretAccessKey': 'foo'}}]:
             with self.assertRaisesRegex(
-                    ValueError, "kms_providers\['aws'\] must contain "
+                    ValueError, r"kms_providers\['aws'\] must contain "
                                 "'accessKeyId' and 'secretAccessKey'"):
                 MongoCryptOptions(invalid_kms_providers)
         with self.assertRaisesRegex(
-                TypeError, "kms_providers\['local'\]\['key'\] must be an "
-                           "instance of bytes or str \(unicode in Python 2\)"):
+                TypeError, r"kms_providers\['local'\]\['key'\] must be an "
+                           r"instance of bytes or str \(unicode in Python 2\)"):
             MongoCryptOptions({'local': {'key': None}})
         with self.assertRaisesRegex(
-                TypeError, "kms_providers\['gcp'\]\['privateKey'\] must be an "
-                           "instance of bytes or str \(unicode in Python 2\)"):
+                TypeError, r"kms_providers\['gcp'\]\['privateKey'\] must be an "
+                           r"instance of bytes or str \(unicode in Python 2\)"):
             MongoCryptOptions({'gcp': {'email': "foo@bar.baz",
                                        "privateKey": None}})
 
@@ -245,7 +255,7 @@ class TestMongoCrypt(unittest.TestCase):
         with km_contexts[0] as kms_ctx:
             self.assertEqual(kms_ctx.kms_provider, "aws")
             self.assertEqual(kms_ctx.endpoint, "kms.us-east-1.amazonaws.com:443")
-            self.assertEqual(len(kms_ctx.message), 781)
+            self.assertEqual(len(kms_ctx.message), 790)
             self.assertEqual(kms_ctx.bytes_needed, 1024)
 
             kms_ctx.feed(http_data('kms-reply.txt'))
@@ -430,6 +440,175 @@ class TestMongoCryptCallback(unittest.TestCase):
                          json_data('command-reply.json'))
         self.assertEqual(decrypted, bson_data('command-reply.json'))
 
+    def test_need_kms_aws_credentials(self):
+        kms_providers = { 'aws': {} }
+        opts = MongoCryptOptions(kms_providers)
+        callback = MockCallback(
+            list_colls_result=bson_data('collection-info.json'),
+            mongocryptd_reply=bson_data('mongocryptd-reply.json'),
+            key_docs=[bson_data('key-document.json')],
+            kms_reply=http_data('kms-reply.txt'))
+        encrypter = AutoEncrypter(callback, opts)
+        self.addCleanup(encrypter.close)
+
+        with mock.patch("pymongocrypt.credentials.aws_temp_credentials") as m:
+            m.return_value = AwsCredential("example", "example", None)
+            decrypted = encrypter.decrypt(
+                bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+
+        self.assertEqual(bson.decode(decrypted, OPTS),
+                         json_data('command-reply.json'))
+        self.assertEqual(decrypted, bson_data('command-reply.json'))
+
+    def test_need_kms_gcp_credentials(self):
+        kms_providers = { 'gcp': {} }
+        opts = MongoCryptOptions(kms_providers)
+        callback = MockCallback(
+            list_colls_result=bson_data('collection-info.json'),
+            mongocryptd_reply=bson_data('mongocryptd-reply.json'),
+            key_docs=[bson_data('key-document-gcp.json')],
+            kms_reply=http_data('kms-reply-gcp.txt'))
+        encrypter = AutoEncrypter(callback, opts)
+        self.addCleanup(encrypter.close)
+
+        with requests_mock.Mocker() as m:
+            data = {"access_token": "foo"}
+            url = "http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token"
+            m.get(url, text=json.dumps(data))
+            decrypted = encrypter.decrypt(
+                bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+
+        self.assertEqual(bson.decode(decrypted, OPTS),
+                         json_data('command-reply.json'))
+        self.assertEqual(decrypted, bson_data('command-reply.json'))
+
+
+class TestNeedKMSAzureCredentials(unittest.TestCase):
+    maxDiff = None
+
+    def get_encrypter(self, clear_cache=True):
+        if clear_cache:
+            pymongocrypt.credentials._azure_creds_cache = None
+        kms_providers = { 'azure': {} }
+        opts = MongoCryptOptions(kms_providers)
+        callback = MockCallback(
+            list_colls_result=bson_data('collection-info.json'),
+            mongocryptd_reply=bson_data('mongocryptd-reply.json'),
+            key_docs=[bson_data('key-document-azure.json')],
+            kms_reply=http_data('kms-reply-azure.txt'))
+        encrypter = AutoEncrypter(callback, opts)
+        self.addCleanup(encrypter.close)
+        return encrypter
+
+    def test_success(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            data = {"access_token": "foo", "expires_in": 4000}
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, text=json.dumps(data))
+            decrypted = encrypter.decrypt(
+                bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+
+        self.assertEqual(bson.decode(decrypted, OPTS),
+                         json_data('command-reply.json'))
+        self.assertEqual(decrypted, bson_data('command-reply.json'))
+        self.assertIsNotNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_empty_json(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, text=json.dumps({}))
+            with self.assertRaisesRegex(MongoCryptError, "Azure IMDS response must contain"):
+                encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+        self.assertIsNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_bad_json(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, text="a'")
+            with self.assertRaisesRegex(MongoCryptError, "Azure IMDS response must be in JSON format"):
+                encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+        self.assertIsNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_http_404(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, status_code=404)
+            with self.assertRaisesRegex(MongoCryptError, "Failed to acquire IMDS access token."):
+                encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+        self.assertIsNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_http_500(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, status_code=500)
+            with self.assertRaisesRegex(MongoCryptError, "Failed to acquire IMDS access token."):
+                encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+        self.assertIsNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_slow_response(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, exc=requests.exceptions.ConnectTimeout)
+            with self.assertRaisesRegex(MongoCryptError, "Failed to acquire IMDS access token: "):
+                encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+        self.assertIsNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_cache(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            data = {"access_token": "foo", "expires_in": 4000}
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, text=json.dumps(data))
+            decrypted = encrypter.decrypt(
+                bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+
+        self.assertIsNotNone(pymongocrypt.credentials._azure_creds_cache)
+
+        # Should use the cached value.
+        decrypted = encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+        self.assertEqual(decrypted, bson_data('command-reply.json'))
+
+        self.assertIsNotNone(pymongocrypt.credentials._azure_creds_cache)
+
+    def test_cache_expires_soon(self):
+        encrypter = self.get_encrypter()
+        with requests_mock.Mocker() as m:
+            data = {"access_token": "foo", "expires_in": 10}
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, text=json.dumps(data))
+            decrypted = encrypter.decrypt(
+                bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+
+        self.assertIsNotNone(pymongocrypt.credentials._azure_creds_cache)
+
+        # Should not use the cached value.
+        encrypter = self.get_encrypter(False)
+        self.assertIsNotNone(pymongocrypt.credentials._azure_creds_cache)
+        with requests_mock.Mocker() as m:
+            url = "http://169.254.169.254/metadata/identity/oauth2/token"
+            m.get(url, exc=requests.exceptions.ConnectTimeout)
+            with self.assertRaisesRegex(MongoCryptError, "Failed to acquire IMDS access token: "):
+                encrypter.decrypt(bson_data('encrypted-command-reply.json'))
+            self.assertTrue(m.called)
+
+        self.assertIsNone(pymongocrypt.credentials._azure_creds_cache)
+
 
 class KeyVaultCallback(MockCallback):
     def __init__(self, kms_reply=None):
